/*
->================================================================================<-
->= JoyDivision - USB joystick adapter - (c) Copyright 2016 OnyxSoft             =<-
->================================================================================<-
->= Version  : 0.1                                                               =<-
->= File     : main.c                                                            =<-
->= Author   : Stefan Blixth (stefan@onyxsoft.se)                                =<-
->= Compiled : 2016-04-25                                                        =<-
->================================================================================<-
->=                                                                              =<-
->= This file is part of JoyDivision - USB joystick adapter                      =<-
->=                                                                              =<-
->= JoyDivision is free software; you can redistribute it and/or modify          =<-
->= it under the terms of the GNU General Public License as published by         =<-
->= the Free Software Foundation; either version 2 of the License, or            =<-
->= (at your option) any later version.                                          =<-
->=                                                                              =<-
->= This program is distributed in the hope that it will be useful,              =<-
->= but WITHOUT ANY WARRANTY; without even the implied warranty of               =<-
->= MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                =<-
->= GNU General Public License for more details.                                 =<-
->=                                                                              =<-
->= You should have received a copy of the GNU General Public License along      =<-
->= with this program; if not, write to the Free Software Foundation, Inc.,      =<-
->= 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.                  =<-
->=                                                                              =<-
->= This project is based on V-USB code from Objective Development Software GmbH =<-
->= http://www.obdev.at/products/vusb/                                           =<-
->=                                                                              =<-
->= Partial based on code/ideas from Raphal Assnat                             =<-
->= http://www.raphnet.net/                                                      =<-
->=                                                                              =<-
->================================================================================<-
*/

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/pgmspace.h>

#include "usbdrv.h"
#include "oddebug.h"


const char usbHidReportDescriptor[50] PROGMEM =
{
    // Joystick
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x04,                    // USAGE (Joystick)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x09, 0x04,                    //   USAGE (Pointer)
    0xa1, 0x00,                    //   COLLECTION (Physical)

    // Buttons
    0x05, 0x09,                    //     USAGE_PAGE (Button)
    0x19, 0x01,                    //     USAGE_MINIMUM (Button 1)
    0x29, 0x03,                    //     USAGE_MAXIMUM (Button 3)
    0x15, 0x00,                    //     LOGICAL_MINIMUM (0)
    0x25, 0x01,                    //     LOGICAL_MAXIMUM (1)
    0x95, 0x03,                    //     REPORT_COUNT (3)
    0x75, 0x01,                    //     REPORT_SIZE (1)
    0x81, 0x02,                    //     INPUT (Data,Var,Abs)
    0x95, 0x01,                    //     REPORT_COUNT (1)
    0x75, 0x05,                    //     REPORT_SIZE (5)
    0x81, 0x03,                    //     INPUT (Constant,Var,Abs)

    // Axis
    0x05, 0x01,                    //     USAGE_PAGE (Generic Desktop)
    0x09, 0x30,                    //     USAGE (X)
    0x09, 0x31,                    //     USAGE (Y)
    0x15, 0x81,                    //     LOGICAL_MINIMUM (-127)
    0x25, 0x7f,                    //     LOGICAL_MAXIMUM (127)
    0x75, 0x08,                    //     REPORT_SIZE (8)
    0x95, 0x02,                    //     REPORT_COUNT (2)
    0x81, 0x02,                    //     INPUT (Data,Var,Abs)

    0xc0,                          //   END_COLLECTION
    0xc0                           // END_COLLECTION
};

typedef struct
{
   uchar buttons;
   char  axis_x;
   char  axis_y;
}report_t;

static report_t reportBuffer;
static uchar  idleRate;                // repeat rate for keyboards, never used for mice
static uchar  currState = 0;           // Keeps track on the current state


uchar usbFunctionSetup(uchar data[8])
{
   usbRequest_t *rq = (void *)data;

   if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS)
   {
      if(rq->bRequest == USBRQ_HID_GET_REPORT)
      {
         usbMsgPtr = (usbMsgPtr_t )&reportBuffer;
         return sizeof(reportBuffer);
      }
      else if(rq->bRequest == USBRQ_HID_GET_IDLE)
      {
         usbMsgPtr = (usbMsgPtr_t)&idleRate;
         return 1;
      }
      else if(rq->bRequest == USBRQ_HID_SET_IDLE)
      {
         idleRate = rq->wValue.bytes[1];
      }
   }

   return 0;
}


static void hardwareInit(void)
{
   /*      1 _____________ 5
             \ o o o o o /
              \ o o o o /
            6   9

   Pin   Mouse         Joystick     AVR-Pin
   1     V-Pulse       Up           PC5
   2     H-Pulse       Down         PC4
   3     VQ-Pulse      Left         PC3
   4     HQ-Pulse      Right        PC2
   5     Button3 (M)   n/c          PC0
   6     Button1 (L)   Button1      PC1
   7     +5V           +5V          +5V
   8     GND           GND          GND
   9     Button2 (R)   Button2      PB0   */

   // Init PortB as input with pullup (except bit 6&7)
   DDRB  = 0x00;
   PORTB = 0b00111111;

   // Init PortC as input with pullup (except bit 6&7)
   DDRC  = 0x00;
   PORTC = 0b00111111;

   // Init PortD as input except the USB-data pins.
   DDRD  = 0x00;
   PORTD = 0b11111001;
}


static void buildReport(void)
{
   uchar tmp;
   uchar x=0, y=0;
   uchar b1=0, b2=0, b3=0;

   currState = PINC;
   tmp = currState^0xff;

   // Read the status of the axis...
   if (tmp & (1<<PC2)) {x = 0x7f;}
   if (tmp & (1<<PC3)) {x = 0x81;}
   if (tmp & (1<<PC4)) {y = 0x7f;}
   if (tmp & (1<<PC5)) {y = 0x81;}

   reportBuffer.axis_x = x;
   reportBuffer.axis_y = y;

   // Read the status of the buttons...
   if (tmp & (1<<PC1)) {b1 = 1;}
   if (tmp & (1<<PC0)) {b3 = 1;}

   currState = PINB;
   tmp = currState^0xff;

   if (tmp & (1<<PB0)) {b2 = 1;}

   // Build mask of the active buttons...
   reportBuffer.buttons = 0x00;
   if (b1) reportBuffer.buttons |= 0x01;
   if (b2) reportBuffer.buttons |= 0x02;
   if (b3) reportBuffer.buttons |= 0x04;
}


int __attribute__((noreturn)) main(void)
{
   uchar i;
   wdt_enable(WDTO_1S);

   odDebugInit();
   hardwareInit();
   usbInit();
   usbDeviceDisconnect();  // enforce re-enumeration, do this while interrupts are disabled!
   i = 0;

   while(--i)
   {                       // fake USB disconnect for > 250 ms
      wdt_reset();
      _delay_ms(1);
   }

   usbDeviceConnect();
   sei();

   for(;;)
   {
      wdt_reset();
      usbPoll();

      if (usbInterruptIsReady())
      {
         buildReport();
         usbSetInterrupt((void *)&reportBuffer, sizeof(reportBuffer));
      }
   }
}

